using System;
using System.Drawing;
using System.Collections;
using System.ComponentModel;
using System.Windows.Forms;
using System.Data;
using System.Diagnostics;
using System.Runtime.InteropServices;
using Tao.OpenGl;

using Tao.Platform.Windows;
using System.Threading;
using System.Drawing.Imaging;
using System.IO;

namespace testellipse
{
    public partial class Form1 : Form
    {
        private static IntPtr hDC;                    // Device Context
        private static IntPtr hRC;                    // Rendering Context
        private static bool[] keys = new bool[256];   // Reading Keyboard status

        private int SelectedForm = 0;
        // 0: Torus, 1:Ellipsoide

        // torus data
        private int torus_radi = 7;

        // ellipsoide data
        double RX = 1.0;
        double RY = 1.0;
        double RZ = 1.0;
        double EXP1=1.0;
        double EXP2 = 1.0;
        
        float v1 = 0.0f;
        float v2 = 0.0f;

        public Form1()
        {
            InitializeComponent();
            this.SetStyle(ControlStyles.DoubleBuffer, true);
            if (!CreateGLWindow())
                return;
            // set initial values
            comboBox1.SelectedIndex=0;
            comboBox1.Show();
            UpdateControllers();

        }
#region OGLWindow
        public bool CreateGLWindow()
        {
            int pixelFormat; // Holds The Results After Searching For A Match
            GC.Collect();   // Request A Collection
            // This Forces A Swap
            Kernel.SetProcessWorkingSetSize(Process.GetCurrentProcess().Handle, -1, -1);
            Cursor.Show();

            Gdi.PIXELFORMATDESCRIPTOR pfd = new Gdi.PIXELFORMATDESCRIPTOR();
            // pfd is set to express what we want
            pfd.nSize = (short)Marshal.SizeOf(pfd);     // Size Of This Pixel Format Descriptor
            pfd.nVersion = 1;                           // Version Number
            pfd.dwFlags = Gdi.PFD_DRAW_TO_WINDOW |      // Format Must Support Window
                Gdi.PFD_SUPPORT_OPENGL |                // Format Must Support OpenGL
                Gdi.PFD_DOUBLEBUFFER;                   // Format Must Support Double Buffering
            pfd.iPixelType = (byte)Gdi.PFD_TYPE_RGBA;   // Request An RGBA Format
            pfd.cColorBits = (byte)16;                  // Select Our Color Depth
            pfd.cRedBits = 0;                           // Color Bits Ignored
            pfd.cRedShift = 0;
            pfd.cGreenBits = 0;
            pfd.cGreenShift = 0;
            pfd.cBlueBits = 0;
            pfd.cBlueShift = 0;
            pfd.cAlphaBits = 0;                         // No Alpha Buffer
            pfd.cAlphaShift = 0;                        // Shift Bit Ignored
            pfd.cAccumBits = 0;                         // No Accumulation Buffer
            pfd.cAccumRedBits = 0;                      // Accumulation Bits Ignored
            pfd.cAccumGreenBits = 0;
            pfd.cAccumBlueBits = 0;
            pfd.cAccumAlphaBits = 0;
            pfd.cDepthBits = 16;                        // 16Bit Z-Buffer (Depth Buffer)
            pfd.cStencilBits = 0;                       // No Stencil Buffer
            pfd.cAuxBuffers = 0;                        // No Auxiliary Buffer
            pfd.iLayerType = (byte)Gdi.PFD_MAIN_PLANE;  // Main Drawing Layer
            pfd.bReserved = 0;                          // Reserved
            pfd.dwLayerMask = 0;                        // Layer Masks Ignored
            pfd.dwVisibleMask = 0;
            pfd.dwDamageMask = 0;

            hDC = User.GetDC(OGLPanel.Handle);
            //hDC = User.GetDC(((Form1)this).Handle);
            if (hDC == IntPtr.Zero)
            {
                // The system could not respond to our needs
                KillGLWindow();
                MessageBox.Show("Cannot Create A GL Device Context.", "ERROR",
                    MessageBoxButtons.OK, MessageBoxIcon.Error);
                return false;
            }

            pixelFormat = Gdi.ChoosePixelFormat(hDC, ref pfd);
            if (pixelFormat == 0)
            {
                // Windows could not find A Matching Pixel Format                                                 
                MessageBox.Show("Cannot Find A Suitable PixelFormat.", "ERROR",
                    MessageBoxButtons.OK, MessageBoxIcon.Error);
                return false;
            }

            if (!Gdi.SetPixelFormat(hDC, pixelFormat, ref pfd))
            {
                // WE were not able to To Set The Pixel Format?
                KillGLWindow();
                MessageBox.Show("Cannot Set The PixelFormat.", "ERROR",
                    MessageBoxButtons.OK, MessageBoxIcon.Error);
                return false;
            }

            hRC = Wgl.wglCreateContext(hDC);
            if (hRC == IntPtr.Zero)
            {
                // Not able to get A Rendering Context?
                KillGLWindow();
                MessageBox.Show("Cannot Create A GL Rendering Context.", "ERROR",
                    MessageBoxButtons.OK, MessageBoxIcon.Error);
                return false;
            }

            if (!Wgl.wglMakeCurrent(hDC, hRC))
            {
                // Not able to activate the Rendering Context
                KillGLWindow();
                MessageBox.Show("Cannot Activate The GL Rendering Context.", "ERROR",
                    MessageBoxButtons.OK, MessageBoxIcon.Error);
                return false;
            }
            // Do one time initialization of OpenGL
            if (!InitGL())
            {
                KillGLWindow();
                MessageBox.Show("Initialization Failed.", "ERROR",
                    MessageBoxButtons.OK, MessageBoxIcon.Error);
                return false;
            }
            // this resize will set perspective
            //ReSizeOpenGL(OGLPanel.Left, OGLPanel.Top, OGLPanel.Width, OGLPanel.Height);

            // poohh, it worked
            return true;
        }


        public void KillGLWindow()
        {
            // Do we have a rendering context?
            if (hRC != IntPtr.Zero)
            {
                // can we release it
                if (!Wgl.wglMakeCurrent(IntPtr.Zero, IntPtr.Zero))
                {
                    MessageBox.Show("Release Of DC And RC Failed.", "SHUTDOWN ERROR",
                        MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
                if (!Wgl.wglDeleteContext(hRC))
                {                                // Are We Able To Delete The RC?
                    MessageBox.Show("Release Rendering Context Failed.", "SHUTDOWN ERROR",
                        MessageBoxButtons.OK, MessageBoxIcon.Error);
                }

                hRC = IntPtr.Zero;
            }

            // Do we have a Device Context?
            if (hDC != IntPtr.Zero)
            {
               /*
                // and a window?
                if (theForm != null && !theForm.IsDisposed)
                {
                    // and a handle to the DC?
                    if (theForm.Handle != IntPtr.Zero)
                    {
                        // Can we release it
                        if (!User.ReleaseDC(theForm.Handle, hDC))
                        {
                            MessageBox.Show("Release Device Context Failed.", "SHUTDOWN ERROR",
                                MessageBoxButtons.OK, MessageBoxIcon.Error);
                        }
                    }
                }
                */
                hDC = IntPtr.Zero;
            }
         }
#endregion OGLWindow

#region initOpenGL
        public  bool InitGL()
        {
            // set up material
            float[] mat_ambient ={ 0.3f, 0.3f, 0.0f, 1.0f };
            float[] mat_diffuse ={ 1.0f, 1.0f, 0.5f, 1.0f };
            float[] spec ={ 0.6f, 0.6f, 0.5f, 1.0f };
            Gl.glMaterialfv(Gl.GL_FRONT, Gl.GL_AMBIENT, mat_ambient);
            Gl.glMaterialfv(Gl.GL_FRONT, Gl.GL_DIFFUSE, mat_diffuse);

            // set up light
            float[] ambient = { 0.2f, 0.2f, 0.2f, 1.0f };
            float[] diffuse = { 1.0f, 1.0f, 1.0f, 1.0f };
            float[] position = { 200.0f, 300.0f, 100.0f, 0.0f };
            Gl.glLightfv(Gl.GL_LIGHT0, Gl.GL_AMBIENT, ambient);
            Gl.glLightfv(Gl.GL_LIGHT0, Gl.GL_DIFFUSE, diffuse);
            Gl.glLightfv(Gl.GL_LIGHT0, Gl.GL_POSITION, position);
            Gl.glEnable(Gl.GL_LIGHT0);
            Gl.glEnable(Gl.GL_LIGHTING);

            // enable light 
            Gl.glEnable(Gl.GL_LIGHTING);
            Gl.glEnable(Gl.GL_LIGHT0);

            // depth sorting
            Gl.glEnable(Gl.GL_DEPTH_TEST);
            Gl.glDepthFunc(Gl.GL_LESS);

            // auto normals
            Gl.glEnable(Gl.GL_NORMALIZE);

            // smooth
            //Gl.glShadeModel(Gl.GL_SMOOTH);

            // background
            Gl.glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
            return true;
        }
#endregion initOpenGL
        

#region Display
        private void Display()
        {
            Gl.glMatrixMode(Gl.GL_PROJECTION);
            Gl.glLoadIdentity();
            int left = OGLPanel.Left;
            int top = OGLPanel.Top;
            int width = OGLPanel.Width;
            int height = OGLPanel.Height;
            Gl.glViewport(0,0,width,height);            
            Glu.gluPerspective(30.0f, (double)width / height, 0.1f, 70.0f);
            
            //---------
            Gl.glMatrixMode(Gl.GL_MODELVIEW);

            Gl.glClear(Gl.GL_COLOR_BUFFER_BIT |
                        Gl.GL_DEPTH_BUFFER_BIT |
                        Gl.GL_ACCUM_BUFFER_BIT);
            Gl.glMatrixMode(Gl.GL_MODELVIEW);

            Gl.glLoadIdentity();
            Glu.gluLookAt(0, 0, 5, // eyepos
                          0, 0, 0, // look at
                          0, 1, 0);// up-vector
            Gl.glRotatef(v1, 1.0f, 0.0f, 0.0f);
            Gl.glRotatef(v2, 0.0f, 0.0f, 1.0f);

            if (SelectedForm == 0)
            {
                float r2=1.0f/(1.0f*torus_radi);
                DrawTorus(1.0f, r2, 50, 50, Gl.GL_TRIANGLE_STRIP);
            }
            else if (SelectedForm == 1)
            {
                DrawSuperEllipsoid(RX, RY, RZ, 50, EXP1, EXP2, Gl.GL_TRIANGLE_STRIP);
            }
            else
                DrawQSphere(Gl.GL_TRIANGLE_STRIP);
            Gdi.SwapBuffers(hDC);
        }
#endregion Display
#region torus
        private static void DrawTorus(float R, float r, int N, int n, int mode)
        {
            //	use parametric expression
            //	z=r.sin(v)
            //	y=(R+rcos(v))sin(w)
            //	x=(R+rcos(v))cos(w)
            //	v running around body of torus
            //	w running round the hole
            //
            //	Strategy:
            //	Assume a larger torus with large radius rr=1.5*r
            //  Normal at any point should be from small to large
            //	Run v as inner loop

            int maxn = 1000; // max precision
            float pi = 3.14159265359f;
            n = Math.Min(n, maxn - 1);
            N = Math.Min(N, maxn - 1);
            float rr = 1.5f * r;

            double dv = 2 * pi / n;
            double dw = 2 * pi / N;

            double v = 0.0f;
            double w = 0.0f;

            // outer loop
            while (w < 2 * pi + dw)
            {
                v = 0.0f;
                Gl.glBegin(mode);
                while (v < 2 * pi + dv)
                {
                    Gl.glNormal3d((R + rr * Math.Cos(v)) * Math.Cos(w) - (R + r * Math.Cos(v)) * Math.Cos(w),
                               (R + rr * Math.Cos(v)) * Math.Sin(w) - (R + r * Math.Cos(v)) * Math.Sin(w),
                                rr * Math.Sin(v) - r * Math.Sin(v));
                    Gl.glVertex3d((R + r * Math.Cos(v)) * Math.Cos(w),
                               (R + r * Math.Cos(v)) * Math.Sin(w),
                                r * Math.Sin(v));

                    Gl.glNormal3d((R + rr * Math.Cos(v + dv)) * Math.Cos(w + dw) - (R + r * Math.Cos(v + dv)) * Math.Cos(w + dw),
                               (R + rr * Math.Cos(v + dv)) * Math.Sin(w + dw) - (R + r * Math.Cos(v + dv)) * Math.Sin(w + dw),
                                rr * Math.Sin(v + dv) - r * Math.Sin(v + dv));
                    Gl.glVertex3d((R + r * Math.Cos(v + dv)) * Math.Cos(w + dw),
                               (R + r * Math.Cos(v + dv)) * Math.Sin(w + dw),
                                r * Math.Sin(v + dv));

                    v += dv;
                } // inner loop
                Gl.glEnd();
                w += dw;
            } //outer loop
        }
#endregion torus
#region qsphere
        private static void DrawQSphere(int mode)
        {
            Glu.GLUquadric q;
            q = Glu.gluNewQuadric(); // note this
            Glu.gluQuadricDrawStyle(q, mode);
            Glu.gluSphere(q, 1.0, 20, 20);
            Glu.gluDeleteQuadric(q);
        }
#endregion qsphere
#region ellipsoid
        private static void DrawSuperEllipsoid(double Rx, double Ry, double Rz,
                                            int N,
                                            double exp1, double exp2,
                                            int mode)
        {
            //Rx,Ry,Rz - radi
            //N - resolution
            //exp1,exp2 - exponentiation of cos and sin of angles
            //mode - drawmode: GL_QUAD_STRIP or GL_TRIANGLE_STRIP
            double pi = Math.PI;
            N = Math.Min(N, 100);// max precision


            // v:[-pi/2,pi/2]
            double dv = pi / N;
            // w:[-pi,pi]
            double dw = 2 * pi / N;


            // outer loop
            double w = -pi;
            for (int wix = 0; wix <= N; wix++)
            {
                Gl.glBegin(mode);
                double v = -pi / 2.0;
                for (int vix = 0; vix <= N; vix++)
                {
                    Point3d p0 = FindPointOnEllipsoid(v, w, exp1, exp2, Rx, Ry, Rz);
                    Point3d p1 = FindPointOnEllipsoid(v + dv, w, exp1, exp2, Rx, Ry, Rz);
                    Point3d p2 = FindPointOnEllipsoid(v + dv, w + dw, exp1, exp2, Rx, Ry, Rz);
                    Point3d p3 = FindPointOnEllipsoid(v, w + dw, exp1, exp2, Rx, Ry, Rz);

                    
                    // normals according to math
                    Point3d n0 = FindNormal(v,      w,       exp1, exp2, Rx, Ry, Rz);
                    Point3d n1 = FindNormal(v + dv, w,       exp1, exp2, Rx, Ry, Rz);
                    Point3d n2 = FindNormal(v + dv, w + dw,  exp1, exp2, Rx, Ry, Rz);
                    Point3d n3 = FindNormal(v,      w + dw,  exp1, exp2, Rx, Ry, Rz);
                    

                    
                    // fake with circle-normals
                    /*Point3d n0 = new Point3d(p0.X, p0.Y, p0.Z);
                    Point3d n1 = new Point3d(p1.X, p1.Y, p1.Z);
                    Point3d n2 = new Point3d(p2.X, p2.Y, p2.Z);
                    Point3d n3 = new Point3d(p3.X, p3.Y, p3.Z);
                      */

                    
                    // with cross product normals, based on quad
                    /*Point3d n0 = FindXNormal(p0, p1, p3);
                    Point3d n1 = FindXNormal(p1, p2, p0);
                    Point3d n2 = FindXNormal(p2, p3, p1);
                    Point3d n3 = FindXNormal(p3, p0, p2);
                    */

                    // draw the squad with a normal in each corner
                    Gl.glNormal3d(n0.X, n0.Y, n0.Z);
                    Gl.glVertex3d(p0.X, p0.Y, p0.Z);

                    /*
                    Gl.glNormal3d(n1.X, n1.Y, n1.Z);
                    Gl.glVertex3d(p1.X, p1.Y, p1.Z);
                    
                    Gl.glNormal3d(n2.X , n2.Y , n2.Z);
                    Gl.glVertex3d(p2.X, p2.Y, p2.Z);
                    */

                    Gl.glNormal3d(n3.X, n3.Y, n3.Z);
                    Gl.glVertex3d(p3.X, p3.Y, p3.Z);

                    v += dv;

                } // inner loop
                Gl.glEnd();
                w += dw;
            } //outer loop
        }
        private static Point3d FindPointOnEllipsoid(double v1, double v2,
                                                    double exp1, double exp2,
                                                    double rx, double ry, double rz)
        {
            double cv1 = Math.Cos(v1);
            double cv2 = Math.Cos(v2);
            double sv1 = Math.Sin(v1);
            double sv2 = Math.Sin(v2);

            double tmp = Math.Sign(cv1) * Math.Pow(Math.Abs(cv1), exp1);
            return new Point3d(rx * tmp * Math.Sign(cv2) * Math.Pow(Math.Abs(cv2), exp2),
                               ry * Math.Sign(sv1) * Math.Pow(Math.Abs(sv1), exp1),
                               rz * tmp * Math.Sign(sv2) * Math.Pow(Math.Abs(sv2), exp2));
        }
        private static Point3d FindNormal(double v1, double v2,
                                         double exp1, double exp2,
                                         double rx, double ry, double rz)
        {
            // e can calculate the normal at any point on a superellipsoid
            double cv1 = Math.Cos(v1);
            double cv2 = Math.Cos(v2);
            double sv1 = Math.Sin(v1);
            double sv2 = Math.Sin(v2);
            return new Point3d(
                 Math.Sign(cv1) * Math.Pow(Math.Abs(cv1), 2.0 - exp1)
                     * Math.Sign(cv2) * Math.Pow(Math.Abs(cv2), 2.0 - exp2) / rx
                 ,
                 Math.Sign(cv1) * Math.Pow(Math.Abs(cv1), 2.0 - exp1)
                     * Math.Sign(sv2) * Math.Pow(Math.Abs(sv2), 2.0 - exp2) / ry
                 ,
                 Math.Sign(sv1) * Math.Pow(Math.Abs(sv1), 2.0 - exp1) / rz
                     );
        }
        private static Point3d FindXNormal(Point3d p0, Point3d p1, Point3d p2)
        {
            // By X-product
            // C=AxB=(a2b3-a3b2, a3b1-a1b3, a1b2-a2b1)
            Point3d a = new Point3d(p1.X - p0.X, p1.Y - p0.Y, p1.Z - p0.Z);
            Point3d b = new Point3d(p2.X - p0.X, p2.Y - p0.Y, p2.Z - p0.Z);
            return new Point3d(
                 a.Y * b.Z - a.Z * b.Y,
                 a.Z * b.X - a.X * b.Z,
                 a.X * b.Y - a.Y * b.X);
        }

#endregion ellipsoid
#region events
        private void Form1_Load(object sender, EventArgs e)
        {
        }
 
        private void OGLPanel_Paint(object sender, PaintEventArgs e)
        {
            Display();
        }

        private void hScrollBar1_Scroll(object sender, ScrollEventArgs e)
        {
            v2=e.NewValue;
            Display();
        }

 
        private void vScrollBar1_Scroll(object sender, ScrollEventArgs e)
        {
            v1 = e.NewValue;
            Display();
         }

        private void comboBox1_SelectedIndexChanged(object sender, EventArgs e)
        {
            SelectedForm = comboBox1.SelectedIndex;
            panelEllipsoide.Visible = false;
            panelTorus.Visible = false;
            if (SelectedForm == 0)
            {
                panelTorus.Visible = true;
            }
            else if (SelectedForm == 1)
            {
                panelEllipsoide.Visible = true;
            }
            Display();
        }

        private void trackBarTorusRadi_Scroll(object sender, EventArgs e)
        {
            torus_radi = trackBarTorusRadi.Value;
            Display();
        }

        private void vScrollBarRX_Scroll(object sender, ScrollEventArgs e)
        {
            RX = 1.0*e.NewValue/10.0;
            UpdateControllers();
            Display();
        }

        private void vScrollBarRY_Scroll(object sender, ScrollEventArgs e)
        {
            RY = 1.0 * e.NewValue / 10.0;
            UpdateControllers();
            Display();
        }

        private void vScrollBarRZ_Scroll(object sender, ScrollEventArgs e)
        {
            RZ = 1.0 * e.NewValue / 10.0;
            UpdateControllers();
            Display();
        }

        private void hScrollBarExp2_Scroll(object sender, ScrollEventArgs e)
        {
            EXP2 = 1.0 * hScrollBarExp2.Value / 10.0;
            UpdateControllers();
            Display();
        }

        private void hScrollBarExp1_Scroll(object sender, ScrollEventArgs e)
        {
            EXP1 = 1.0 * hScrollBarExp1.Value / 10.0;
            UpdateControllers();
            Display();
        }

        private void UpdateControllers()
        {
            textBoxExp1.Text = Convert.ToString(EXP1); 
            textBoxExp2.Text = Convert.ToString(EXP2);
            textBoxRX.Text = Convert.ToString(RX); 
            textBoxRY.Text = Convert.ToString(RY);
            textBoxRZ.Text = Convert.ToString(RZ);
        }
#endregion events
    }
#region point
    class Point3d : Object
    {
        double x, y, z;
        public Point3d(double x, double y, double z)
        {
            this.x = x;
            this.y = y;
            this.z = z;
        }
        public double X { get { return x; } set { x = value; } }
        public double Y { get { return y; } set { y = value; } }
        public double Z { get { return z; } set { z = value; } }
        public override string ToString()
        {
            return Convert.ToString(X) + "," + Convert.ToString(Y) + "," + Convert.ToString(Z);
        }
    }
#endregion point
}